#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include "shaunlib.h"

// COMPSOL

// Run EXTRACTSOL on L-POP output first

// Utility to make comparison of different 
// LPOP solutions

const int LARGEST = 50;   // Max # of classes in a solution

int main()
{
  
  vector<string> tag;
  vector<int> cl;
  vector<string> ctag;
  string tag_;
  int cl_;
  int cnt = 1; 

  while (!cin.eof()) 
    {
      cin >> tag_;
      cin >> cl_;
      cin>>ws;
      tag.push_back(tag_);
      cl.push_back(cl_);
      cnt++;
    }

  int clcnt = 1; 
  ctag.push_back(tag[0]);
  for (int i = 1 ; i < tag.size() ; i++) 
    if ( tag[i] != tag[i-1] ) 
      {
	clcnt++;
	ctag.push_back(tag[i]);
      }    
  
  int icnt = cnt / clcnt;

  cout.precision(4);
  cout.setf(ios::fixed);
  cout << "COMPSOL : Compare L-POP solutions \n";
  cout << "\n" << clcnt << " solutions and " << icnt << " individuals.\n";

  CMatrix a(icnt,clcnt);
  IntArray cln;
  cln.setCol(clcnt);

  int k = 0 ; 
  for (int j = 1 ; j <= clcnt ; j++)
    for (int i = 1 ; i <= icnt ; i++)
      a[i][j] = cl[k++];

  
  // Calculate statistics


  IntMatrix lab;
  lab.setRow(clcnt);
  lab.setCol(LARGEST);

  // Get # of non-zero classes in each solution
  for (int j = 1 ; j <= clcnt ; j++)
    {
      IntArray insol;
      insol.setCol(1);
      insol(1) = (int)a[1][j];
      for (int i = 1 ; i <= icnt ; i++)
	{
	  bool matched = false;
	  int n = insol().size(); 
	  for (int k = 1 ; k <= n ; k++)
	    if ( a[i][j] == insol(k) ) matched = true;
	  if (!matched) 
	    {
	      insol.setCol(n+1);
	      insol(n+1) = (int)a[i][j];
	    }
	}
      cln(j) = insol().size();
      for (int ii = 1 ; ii <= cln(j) ; ii++) lab(j)(ii) = insol(ii);
      cout << "\tSolution \"" << ctag[j-1] << "\" has " << cln(j) << " classes.\n";
    }

 
  // For each pair of solutions, compute RAND
  CMatrix rand_orig(clcnt,clcnt);
  CMatrix rand_adj(clcnt,clcnt);

  for (int sol1 = 1 ; sol1 <= clcnt ; sol1++)
    for (int sol2 = 1 ; sol2 <= clcnt ; sol2++)
      {
	///////////////////////////////////////
	// Calculate N contingency matrix
	IntMatrix n;
	n.setRow(cln(sol1));
	n.setCol(cln(sol2));
	for (int i = 1 ; i <= cln(sol1) ; i++)
	  for (int j = 1 ; j <= cln(sol2) ; j++)
	    {
	      int n_count = 0;
	      for (int ind = 1 ; ind <= icnt ; ind++)
		if ( a[ind][sol1] == lab(sol1)(i) 
		     && a[ind][sol2] == lab(sol2)(j) ) 
		  n_count++;
	      n(i)(j) = n_count;
	    }

	

	// Marginals 
	IntArray n1;
	n1.setCol(cln(sol1));
	
	IntArray n2;
	n2.setCol(cln(sol2));

	for (int i = 1 ; i <= cln(sol1) ; i++)
	  {
	    n1(i) = 0 ; 
	    for (int j = 1 ; j <= cln(sol2) ; j++)
	      n1(i) += n(i)(j);
	  }

	for (int j = 1 ; j <= cln(sol2) ; j++)
	  {
	    n2(j) = 0 ; 
	    for (int i = 1 ; i <= cln(sol1) ; i++)
	      n2(j) += n(i)(j);
	  }

	// Total 
	int n_tot = 0 ; 
	for (int i = 1 ; i <= cln(sol1) ; i++)
	  n_tot += n1(i);


	///////////////////////////////////////
	// Display contingency table
	if (sol1 < sol2)
	  {	
	    cout << "\n"<<ctag[sol2-1]<<"\t";
	    for (int j = 1 ; j <= cln(sol2) ; j++)
	      cout << lab(sol2)(j) << "\t";
	    cout <<"\n"<<ctag[sol1-1]<<"\t";
	    for (int j = 1 ; j <= cln(sol2) ; j++)
	      cout << "-------" << "\t";
	    
	    for (int i = 1 ; i <= cln(sol1) ; i++)
	      {
	    cout << "\n" << lab(sol1)(i) << " |" ; 
	    for (int j = 1 ; j <= cln(sol2) ; j++)
	      cout << "\t" << n(i)(j);
	  }
	    cout << "\n";
	  }
	
	///////////////////////////////////////
	// Calculate RAND - Non-adjusted

	double rand_T = 0 ; 
	for (int i = 1 ; i <= cln(sol1) ; i++)
	  for (int j = 1 ; j <= cln(sol2) ; j++)
	    rand_T += mymath::sqr(n(i)(j));
	rand_T -= n_tot;

	double rand_P = 0 ; 
	for (int i = 1 ; i <= cln(sol1) ; i++)
	  rand_P += mymath::sqr(n1(i));
	rand_P -= n_tot;
	
	double rand_Q = 0 ; 
	for (int j = 1 ; j <= cln(sol2) ; j++)
	  rand_Q += mymath::sqr(n2(j));
	rand_Q -= n_tot;

	rand_orig[sol1][sol2] = (rand_T - 0.5*rand_P - 0.5*rand_Q 
			 + mymath::combin(n_tot,2) ) 
                            / mymath::combin(n_tot,2) ;

	  
	///////////////////////////////////////
	// Calculate RAND - Adj
	
	// N

	double rand1 = 0 ; 
	for (int i = 1 ; i <= cln(sol1) ; i++)
	  for (int j = 1 ; j <= cln(sol2) ; j++)
	    rand1 += mymath::combin(n(i)(j),2);
	
	double rand2a = 0;
	for (int i = 1 ; i <= cln(sol1) ; i++)
	  rand2a += mymath::combin(n1(i),2);
	double rand2b = 0;
	for (int j = 1 ; j <= cln(sol2) ; j++)
	  rand2b += mymath::combin(n2(j),2);
	double rand3 = rand2a * rand2b;
	rand3 /= (double)mymath::combin(n_tot,2);
	double rand_N = rand1 - rand3;
	
	// D
	double rand4 = (rand2a + rand2b ) / 2;
	double rand_D = rand4 - rand3;
	
	rand_adj[sol1][sol2] = rand_N / rand_D;

      }



  ///////////////////////////////////////
  // Display RAND coeficients
  cout << "\nRAND coefficients\n\t";
  for (int j = 1 ; j <= clcnt ; j++)
    cout << ctag[j-1] << "\t";
  cout <<"\n\t";
  for (int j = 1 ; j <= clcnt ; j++)
    cout << "-------" << "\t";
  for (int i = 1 ; i <= clcnt ; i++)
    {
      cout << "\n" << ctag[i-1] << " |" ; 
      for (int j = 1 ; j <= clcnt ; j++)
	cout << "\t" << rand_orig[i][j];
    }
  cout << "\n";


  ///////////////////////////////////////
  // Display RAND adj coeficients
  cout << "\nRAND (adj) coefficients\n\t";
  for (int j = 1 ; j <= clcnt ; j++)
    cout << ctag[j-1] << "\t";
  cout <<"\n\t";  
  for (int j = 1 ; j <= clcnt ; j++)
    cout << "-------" << "\t";
  for (int i = 1 ; i <= clcnt ; i++)
    {
      cout << "\n" << ctag[i-1] << " |" ; 
      for (int j = 1 ; j <= clcnt ; j++)
	cout << "\t" << rand_adj[i][j];
    }
  cout << "\n";

}



