#include "data.h"
#include <fstream>

// ----------------------------------------------------------------------------------
// READ IN DATA FILE into d1[][] and d2[][], id[] and p[][]
// ----------------------------------------------------------------------------------

void Data::load(Param &p)
{

  ifstream in; 

  in.open(p.datafile.c_str());

  if (p.header_row) in.ignore(50000,'\n');
  
  for(int i = 1; i <= nInd; i++) 
    {
    

  // In L-POP file format
      if (!p.structure_format) 
	{
	  in >> id(i);                           

	  if (p.fix_to_class) in >> fix2c(i);

// IGNORE PHENOTYPE DATA FOR NOW
// AND ALLOW TO BE STRING FORMAT
	  string dustbin;
	  for(int j=1; j <= p.nPheno; j++)
	    in >> dustbin;                    

	  if (!p.haploid) // DIPLOID
	    for(int j = 1; j <= p.nLoci; j++)            // for each loci
	      in >> d1(i)(j) >> d2(i)(j);     // get the 2 alleles

	  if (p.haploid) // HAPLOID
	    for (int j = 1; j <= p.nLoci; j++)
	      {
		in >> d1(i)(j);
		d2(i)(j) = d1(i)(j);
	      }

	  // Make missing if FIX-TO-CLASS is -1 
	  if (p.fix_to_class && fix2c(i)==-1) 
	    for (int j = 1; j <= p.nLoci; j++) 
	      d1(i)(j) = d2(i)(j) = p.valMissing;
	}

      // In STRUCTURE file format
      if (p.structure_format)
	{
	  in >> id(i);
	  if (p.fix_to_class) in >> fix2c(i);

	  string dustbin;
	  for(int j=1; j <= p.nPheno; j++)
	    in >> dustbin;                    

	  for(int j = 1; j <= p.nLoci; j++)            // for each loci
	    in >> d1(i)(j);
	  
	  // Repeat for next haploid line to get
	  // if diploid data 
	  if (!p.haploid) // DIPLOID 
	    {
	      in >> id(i); 
	      if (p.fix_to_class) in >> fix2c(i);
	      for(int j=1; j <= p.nPheno; j++)
		in >> dustbin;                      // and the phenotypes
	      for(int j = 1; j <= p.nLoci; j++)            // for each loci
		in >> d2(i)(j);
	    }

	  if (p.haploid) // HAPLOID
	    for (int j=1 ; j <= p.nLoci; j++)
	      d2(i)(j) = d1(i)(j);

	  // Make missing if FIX-TO-CLASS is -1 
	  if (p.fix_to_class && fix2c(i)==-1) 
	    for (int j = 1; j <= p.nLoci; j++) 
	      d1(i)(j) = d2(i)(j) = p.valMissing;
	}

      // Set flag if any X data detected
      for (int loc = 1 ; loc <= nLoci ; loc++)
	if (d2(i)(loc) == p.valX || d1(i)(loc) == p.valX ) p.xdata = true;
	  
    }

  in.close();
  

}





void Data::process(Param &p)
{

  //////////////////////////////////////////
  // If a genotype has one missing allele, 
  // the other allele is coded as missing 
  
  for (int i=1; i<=nInd; i++)
    for (int l=1; l<=p.nLoci; l++)
      if (d1(i)(l) == p.valMissing) { d2(i)(l) = p.valMissing; }
      else if (d2(i)(l) == p.valMissing) { d1(i)(l) = p.valMissing; }


  //////////////////////////////////////////
  // Put any X-chr status in slot 1

  for (int l=1; l<=p.nLoci; l++)    
    for (int i=1; i<=nInd; i++)
      {
	if(d1(i)(l) == p.valX ) 
	  {
	    d1(i)(l) = d2(i)(l);
	    d2(i)(l) = p.valX;
	  }
      }

  /////////////////////////////////////////
  // Encode configurations into c1[ind][locus]
  // and c2[i][l] 

  ////////////////////////////////////////////
  // create allele codes (equated within loci)

  ///////////////////////////////////////////////////
  // Enter first stack entry as all missing values

  for (int lo=1; lo <= p.nLoci; lo++)
    {
      // Set up missing allele code '1'

      // c[locus][code] = missing value code
      c(lo)(1) = p.valMissing;
      fallele(lo)(1) = 0;
      nallele(lo) = 1;

      // Similar procedure for Xchr code '2'
      c(lo)(2) = p.valX;
      fallele(lo)(2) = 0;
      nallele(lo) = 2;
      
    } // end of l, looping over all loci
  
    // set total number of s_configs to 0;
    nconfig = 0;


    // **********************************************************
    // *** Loop for entire sample to build up locus-specific
    // *** coding system
    // **********************************************************

   
    for (int i=1; i <= nInd; i++)         // loop over all individuals
      {                                     // including individual '1';

	for (int l=1; l <= nLoci; l++)    // loop over all loci
	  {

	    /////////////////////
	    // Consider allele 1

	    int ii=1;
	    int iii=0;  // flag variable
	    do
	      {
		// search for entry in reference table c[]

		//match? : return 1
		if (d1(i)(l) == c(l)(ii))
		  iii=1;
		else
		  {
		    ii++;
		    // end of list ? : return 2
		    if (ii > nallele(l) ) iii=2;
		  }
	      } while (iii==0);
	    // increment frequency of existing stack entry
	    if (iii==1) fallele(l)(ii)++;
	    // add new entry to stack
	    if (iii==2)
	      {
		// record actual allele in c[]
		c(l)(ii) = d1(i)(l);
		// set frequency of this allele count to 1
		fallele(l)(ii) = 1;
		// increase the number of alleles at that locus
		nallele(l)++;
	      } // end of 'new entry conditional'
            // assign code to individual's allele
	    s1(i)(l) = ii;
	    
	    ///////////////////////
	    // Consider allele 2
	    
	    ii=1;
	    iii=0;  // flag variable
	    do
	      {
		// search for entry in reference table c[]
		
		//match? : return 1
		if (d2(i)(l) == c(l)(ii))
		  iii=1;
		else
		  {
		    ii++;
		    // end of list ? : return 2
		    if (ii > nallele(l) ) iii=2;
		  }
	      } while (iii==0);
	    // increment frequency of existing stack entry
	    if (iii==1) fallele(l)(ii)++;
	    // add new entry to stack
	    if (iii==2)
	      {
		// record actual allele in c[]
		c(l)(ii) = d2(i)(l);
		// set frequency of this allele count to 1
		fallele(l)(ii) = 1;
		// increase the number of alleles at that locus
		nallele(l)++;
	      } // end of 'new entry conditional'
	    // assign code to individual's allele
	    s2(i)(l) = ii;
	    
	    
	    
	    ////////////////////////
	    // 
	    
	    // Swap alleles in s[] if necessary
	    
	    if (s2(i)(l) != 2)
	      if (s1(i)(l) > s2(i)(l))
		{
		  int temp = s1(i)(l);
		  s1(i)(l) = s2(i)(l);
		  s2(i)(l) = temp;
		}

	    // check that we haven't gone over the ALLELE count
	    if (nallele(l) > p.nAllele) error("Need to specify higher ALLELE count");
	    
	  } // end of 'l' loop
      } // end of 'i' loop
    


    // **************************************************
    // *** Construct configuration coding system
    // *** i.e. not locus-specific
    // *** but based on recoded dataset s[] instead of d[]
    // **************************************************


  // ----------------------------------------------------------------------------------
  // GENERATE CONFIG-BASED CODING SYSTEM TO MAKE INDIVIDUALS CONFIGURATIONS
  // ----------------------------------------------------------------------------------

    if (p.group_geno)
      {    
	for (int i=1; i <= nInd; i++)
	  {
	    
	    int ii=1;
	    int iii=0;
	    
	    do
	      {

		//match? : return 1
		iii=1;  // return 1 for a match, unless is set to 0


		if (i>1) // ensure first ind is config 1
		  {
		    for (int l=1; l <= p.nLoci; l++)
		      if ((s1(i)(l) != config1(ii)(l))
			  || (s2(i)(l) != config2(ii)(l))) iii=0;
		  }

		// else increment
		if (iii==0) ii++;

		// end of list ? : return 2
		if (ii>nconfig) iii=2;

	      } while (iii==0);

	    
	    // increment existing stack entry
	    if (iii==1) fconfig(ii)++;
	    
	    // add new entry to stack
	    if (iii==2)
	    {
	      for (int l=1; l <= p.nLoci; l++)
		{
		  config1(ii)(l) = s1(i)(l);
		  config2(ii)(l) = s2(i)(l);
		}
	      
	      fconfig(ii) = 1;
	      nconfig++;
	    }
	    
	    // record individuals configuration
	    ic(i) = ii;
	  }
      }


    // ----------------------------------------------------------------------------------
    // GROUP_GENO OPTION NOT SET - FOR CONSISTENCY WITH REST OF CODE MAKE PRETEND 
    // GROUPING WHERE EACH CONFIGURATION N_i = 1 AND IS NOT UNIQUE - I.E. INDIVIDUALS
    // ----------------------------------------------------------------------------------

    if (!p.group_geno)
      {    
	for (int i=1; i <= nInd; i++)
	  {
	    // add new entry to stack
	    nconfig++;
	    
	    for (int l=1; l <= p.nLoci; l++)
	      {
		config1(nconfig)(l) = s1(i)(l);
		config2(nconfig)(l) = s2(i)(l);
	      }
	    
	    fconfig(nconfig) = 1;
	    
	    // record individuals configuration
	    ic(i) = nconfig;

	  }
      }


//     // ----------------------------------------------------------------------------------
//     // GET MAXIMUM # OF ALLELES AT ANY ONE LOCUS 
//     // ----------------------------------------------------------------------------------

//     for (int loc = 1 ; loc <= nLoci ; loc++)
//       if (nallele(loc) > p.nAllele ) p.nAllele = nallele(loc);

    
    // ----------------------------------------------------------------------------------
    // FIX-TO-CLASS IMPLEMENTATION (THIS MAY BE DANGEROUS IF GROUP_GENO == TRUE
    // ----------------------------------------------------------------------------------
    
    fixconfig2c.setCol(nconfig);
    
    for (int i=1; i <= p.nInd; i++)
      fixconfig2c(ic(i)) = fix2c(i);
}



void Data::hw_recode(Param &p)
{

  // Check no X chr data
  if (p.xdata) error("Cannot use HWD option w/ male X chr data");
  
  // Set to haploid organism
  p.haploid = true;

  for (int i = 1 ; i <= nInd ; i++)
    for (int loc = 1 ; loc <= nLoci ; loc++)
      {
	if (d1(i)(loc) == p.valMissing || d2(i)(loc) == p.valMissing)
	  {
	    d1(i)(loc) = p.valMissing;
	    d2(i)(loc) = p.valX;
	  }
	else
	  {
	    string buf1, buf2; 
	    if (d1(i)(loc) <= d2(i)(loc))
	      {
		buf1 = int2str(d1(i)(loc));
		buf2 = int2str(d2(i)(loc));
	      }
	    else
	      {
		buf1 = int2str(d2(i)(loc));
		buf2 = int2str(d1(i)(loc));
	      }
	    string buf3 = buf1 + buf2;
	    int buf4 = atoi(buf3.c_str());
	    d1(i)(loc) = buf4;
	    d2(i)(loc) = p.valX;
	  }
      }
}
