#include <iostream>
#include <iomanip>
#include <fstream>
#include <string> 
#include <math.h>
#include <stdlib.h>

#include "shaunlib.h"
#include "data.h"
#include "param.h"
#include "lca.h"
#include "output.h"

using namespace std;

int main(int argc, char* argv[])
{

  CArgs a(argc,argv);
  if (a.any()) 
    {
      cerr << "\n"
	   << "Title     : L-POP version 1.2\n"
	   << "Authors   : S. Purcell & P. Sham, Nov. 2001\n"
	   << "Usage     : lpop < {parameter file} > {output file} \n"
	   << "More info : http://statgen.iop.kcl.ac.uk/l-pop/ \n";
      exit(0);
   
    } 
  
  
  // ----------------------------------------------------------------------------------
  // GET PARAMETERS AND DATA
  // ----------------------------------------------------------------------------------


  Data                  data;
  Param                 p;
  LCAParam              lp;

  p.get_param           (lp, data); 
  p.get_loci_ind        (lp,data);
  p.get_class           (lp);
  p.get_storage         (lp,data);
  lp.make_admix         ();
  data.load             (p);

  // If option selected, recode genotypes as alleles
  if (p.hwd) data.hw_recode(p);

  data.process          (p);


  
  // ----------------------------------------------------------------------------------
  // BASIC LCA ANALYSIS
  // ----------------------------------------------------------------------------------
      
  if (!p.distance)
    {

      Output out            (p,lp);
      if (!p.min) out.basic(p,lp,data);
      if (!p.min) out.data(p,data);
      LCA                   (p,lp,data,out);

      // Normal output of estimates
      if (!p.min) out.estimates(p,lp,data);

      // Minimal output
      if (p.min) out.minimal(p,lp);

      // Optional output of 'entropy'
      if (!p.min && p.entropy) out.entropy(p,lp,data);

      // Optional output of locus information
      if (!p.min && p.locinfo) out.locinfo(p,lp,data);

      // Optional output of cslocus information
      if (!p.min && p.cslocinfo) out.cslocinfo(p,lp,data);

      // Genetic distance between classes
      if (!p.min) out.distance(p, lp,data);
      
      return 0;
    }
  else 
    {
      // ----------------------------------------------------------------------------------
      // DISTANCE ANALYSIS
      // ----------------------------------------------------------------------------------
    
      if (p.group_geno)
	error("Cannot group genoptyes in DISTANCE mode.");

      p.verbose1 = true;
      p.verbose2 = false;
      p.verbose3 = false;
      p.min = false;
      p.nItRepeat = 1;     

      lp.free_storage();
      lp.nPureClass = 1; 
      lp.nAncClass = 1;
      lp.nClass = 1;
      lp.nAdmixClass = 0;
      lp.nAdmix2Class = 0;
      p.get_storage(lp,data);
      lp.make_admix();
      
      p.fix_to_class = false; 
      Output out(p,lp);
      LCA(p,lp,data,out);

      lp.free_storage();
      p.verbose1 = false;
      p.verbose2 = false;
      p.verbose3 = false;
      lp.nPureClass = 2; 
      lp.nClass = 2;
      lp.nAncClass = 2;
      lp.nAdmixClass = 0;
      lp.nAdmixClass2 = 0;
      lp.nAdmix2Class = 0;
      p.get_storage(lp,data);
      lp.make_admix();
      p.fix_to_class = true; 
      Output out_d(p,lp);

      cout << "\nD:"<<p.tag<<": Individual distance measures";
      cout << "\nD:"<<p.tag<<":\tID\tD(i)";

      for (int i = 1 ; i <= p.nInd ; i++)
	{

	  // Exclude one indivudal at a time
  	  for (int ii = 1 ; ii <= p.nInd; ii++)
  	    if (i==ii) data.fixconfig2c(ii) = 2;  
  	    else data.fixconfig2c(ii) = 1;

	  LCA(p,lp,data,out_d);

	  cout << "\nD:"<<p.tag<<":\t" 
	       << data.id(i) 
	       << "\t" 
	       << out_d.ml - out.ml ;
	}
 
      cout << "\n";
      exit(0); 
      
    }
}













