#include "output.h"
#include <math.h>

  // ----------------------------------------------------------------------------------
  // DISPLAY INFORMATION ABOUT THE DATA AND THE MODEL
  // ----------------------------------------------------------------------------------

Output::Output(Param &p, LCAParam &lp)
{
  h.setRow(lp.nClass); 
  h.setCol(p.nInd);
  r = mymath::nr_3tensor(1,p.nLoci,1,lp.nClass,1,p.nAllele);
  n.setCol(lp.nClass);
  ml = 0;
  ii = 0;
  nLoci = p.nLoci;
  nClass = lp.nClass;
  nAllele = p.nAllele;
}

Output::~Output()
{
  mymath::free_nr_3tensor(r,1,nLoci,1,nClass,1,nAllele);
}

void Output::basic(Param &p, LCAParam &lp, Data &d)
{
  if (!p.min)
    {
      cout << "++ L-POP Version 1.2 -- Shaun Purcell & Pak Sham -- Nov. 2001 ++ \n";
      
      cout << "\nO:"<<p.tag<<": Parameter details" 
	   << "\nO:"<<p.tag<<": title                      = " << p.title
	   << "\nO:"<<p.tag<<": datafile                   = " << p.datafile 
	   << "\nO:"<<p.tag<<": datafile format            = ";
      if (!p.structure_format) { cout << "L-POP"; } else { cout << "STRUCTURE" ; }
      cout << "\nO:"<<p.tag<<": number of individuals      = " << d.nInd
	   << "\nO:"<<p.tag<<": number of loci             = " << d.nLoci
	   << "\nO:"<<p.tag<<": max number of alleles      = " << p.nAllele
	   << "\nO:"<<p.tag<<": missing allele value       = " << p.valMissing
	   << "\nO:"<<p.tag<<": X chromosome value         = " << p.valX
	   << "\nO:"<<p.tag<<": number of phenotypes       = " << p.nPheno
	   << "\nO:"<<p.tag<<": organism type              = ";
      if (p.haploid==1) {cout << "Haploid"; } else { cout << "Diploid";} 

      // Distance versus normal analysis
      if (p.distance) cout << "\nO:"<<p.tag<< ": distance analysis           = ON";
      else
	{
	  cout << "\nO:"<<p.tag<<": fix-to-class               = ";
	  if (p.fix_to_class) {cout << "ON"; } else { cout << "OFF";} 
	  cout << "\nO:"<<p.tag<<": number of pure classes     = " << lp.nPureClass
	       << "\nO:"<<p.tag<<": number of admix1 classes   = " << lp.nAdmixClass 
	       << "\nO:"<<p.tag<<": admix1 level               = " << p.admix_res
	       << "\nO:"<<p.tag<<": number of admix2 classes   = " << lp.nAdmix2Class
	       << "\nO:"<<p.tag<<": total number of classes    = " << lp.nClass 
	       << "\nO:"<<p.tag<<": group individuals          = " ; 
	  if (p.group_geno) {cout << "ON"; } else { cout << "OFF";} 
	}

      // Rest of options
      cout << "\nO:"<<p.tag<<": number of E-M repeats      = " << p.nItRepeat
	   << "\nO:"<<p.tag<<": E-M tolerance              = " << p.tol
	   << "\nO:"<<p.tag<<": verbosity level 1          = " ; 
      if (p.verbose1) {cout << "ON"; } else { cout << "OFF";} 
      cout << "\nO:"<<p.tag<<": verbosity level 2          = " ;
      if (p.verbose2) {cout << "ON"; } else { cout << "OFF";} 
      cout << "\nO:"<<p.tag<<": verbosity level 3          = " ;
      if (p.verbose3) {cout << "ON"; } else { cout << "OFF";} 
      if (p.tag=="") { cout 
	<< "\nO:"<<p.tag<<": output tag                 = <none>"; }
      else { cout 
	<< "\nO:"<<p.tag<<": output tag                 = " << p.tag; }
      cout << "\nO:"<<p.tag<<": random number seed         = " << p.random_seed
	   << "\n";
    }


  // ----------------------------------------------------------------------------------
  // OUTPUT CLASS STRUCTURE
  // ----------------------------------------------------------------------------------

  if (!p.min)
    {
      cout << "\nX:"<<p.tag<<":\tCls(N)\tType";
      for (int j = 1; j <= lp.nPureClass; j++)
	cout << "\tP" << j;
      for (int j = 1; j <= lp.nAdmixClass; j++)
	cout << "\tA" << j+lp.nPureClass;

      for (int i = 1; i <= lp.nClass; i++)
	{
	  cout << "\nC:"<<p.tag<<":\t" << i;
	  if (i <= lp.nPureClass) cout << "\tPure";
	  if ((i <= lp.nPureClass+(int)lp.nAdmixClass2) && ( i > lp.nPureClass)) cout << "\tAdmix1";
	  if (i > lp.nPureClass+(int)lp.nAdmixClass2) cout << "\tAdmix2";
	  
	  for (int j = 1; j <= lp.nPureClass+lp.nAdmixClass; j++)
	    cout << "\t" << lp.cls[i][j];
	}
      cout << "\n";
    }
}


// ----------------------------------------------------------------------------------
// DISPLAY INFORMATION ABOUT CONFIGURATIONS AND ALLELE FREQUENCIES IN SAMPLE
// ----------------------------------------------------------------------------------

void Output::data(Param &p, Data &d)
{    
  if (!p.min)
    {
      if (p.verbose1)
	{
	  cout << "\nX:"<<p.tag<<": Genotypic configuration information";
	  for (int ii=1; ii<=d.nconfig; ii++)
	    {
	      cout << "\nG:"<<p.tag<<": " << ii << " N=" << d.fconfig(ii) << "   ";
	      for (int l=1; l<=d.nLoci; l++)
		if (!p.haploid)
		  cout << d.c(l)(d.config1(ii)(l)) << "/" <<
		    d.c(l)(d.config2(ii)(l)) << " ";
		else
		  cout << d.c(l)(d.config1(ii)(l)) << "/" <<
		    p.valX << " ";		  
	    }
	  
	  cout << "\n\nX:"<<p.tag<<": Allele frequency information";
	  
	  for (int l=1; l<= d.nLoci; l++)
	    {
	      cout << "\nA:"<<p.tag<<": " << l  << "\t";
	      for (int a=1; a<=d.nallele(l); a++)
		{
		  if (a!=2) // do not print Xchr codes
		    {
		      // Calculate total # of alleles for that locus
		      int denom = 0;
		      for (int i=1 ; i <= d.nInd; i++)
			if (d.config2(i)(l)==2) denom += 1; 
			else denom += 2; 
		      
		      cout << "P(" << d.c(l)(a) << ")= " << d.fallele(l)(a) 
			   << "(" << (float)d.fallele(l)(a)/denom << ")"
			   << "\t";
		    }
		}
	    }
	  cout << "\n";
	}    
    }
}



// ----------------------------------------------------------------------------------
// OUTPUT FIT-STATISTICS AND PARAMETER ESTIMATES FROM BEST-FIT CLASS
// ----------------------------------------------------------------------------------

void Output::minimal(Param &p, LCAParam &lp)
{
  cout << p.tag << "\t" 
       <<-2*ml << "\t" 
       << lp.npar << "\t" 
       << -(2*ml) + 2*lp.npar 
       << "\n";
}

void Output::estimates(Param &p, LCAParam &lp, Data &d)
{
  cout << "\n\nX:"<<p.tag<<": Fit statistics";
  cout << "\nK:"<<p.tag<<":\tML\t" << -2*ml << "\t";
  cout << "\nK:"<<p.tag<<":\tN(P)\t" << lp.npar << "\t";
  cout << "\nK:"<<p.tag<<":\tAIC\t" << -(2*ml) + 2*lp.npar << "\t";
  cout << "\nK:"<<p.tag<<":\titer\t" << ii << "\t";
  
  // ----------------------------------------------------------------------------------
  // OUTPUT PRIOR LATENT CLASS PROBABILITIES
  // ----------------------------------------------------------------------------------

  cout << "\n\nX:"<<p.tag<<": Prior latent class probabilities P(C)";
  for (int j=1; j<=lp.nClass; j++)
    cout << "\nP:"<<p.tag<<":\t" << j << "\t" << n[j];
  cout << "\n";

  // ----------------------------------------------------------------------------------
  // OUTPUT POSTERIOR LATENT CLASS PROBABILITIES
  // ----------------------------------------------------------------------------------

  cout << "\nX:"<<p.tag<<": Posterior class probabilities P(C|R)";
  cout << "\nX:"<<p.tag<<":\tID\tEst(k)";
  for (int j=1; j<=lp.nClass; j++) cout << "\th("<<j<<"|R)";
  
  for (int i=1; i <= d.nInd; i++)
    {
      cout << "\nI:"<<p.tag<<":\t" << d.id(i);
      
      if (!p.fix_to_class || d.fix2c(i) == 0 ) cout << "\t" << lp.indClass(i);
      if (p.fix_to_class  && d.fix2c(i) > 0 ) cout << "\t[" << lp.indClass(i) << "]";
      if (p.fix_to_class  && d.fix2c(i) == -1 ) cout << "\t[.]";
      
      for (int j=1; j <= lp.nClass; j++)
	cout << "\t" << h[j][d.ic(i)]; 
    }

  // ----------------------------------------------------------------------------------
  // OUTPUT CLASS-SPECIFIC ALLELE FREQUENCIES
  // ----------------------------------------------------------------------------------

  cout << "\n\nX:"<<p.tag<<": Class-specific allele frequencies P(R|C)";
  
  for (int l=1; l <= d.nLoci; l++)
    {
      for (int j=1; j <= lp.nClass; j++)
	{
	  cout << "\nM:"<<p.tag<<":\tl=" << l << "\tj=" << j;
	  for (int k=3; k<=d.nallele(l); k++) // nonmissing alleles only		  
	    cout << "\tp(A="<<d.c(l)(k)<<"|j)= "<< r[l][j][k];
	}
      cout << "\n";
    }
  cout << "\n";
  
}




// ----------------------------------------------------------------------------------
// OPTIONAL OUTPUT OF 'ENTROPY' MEASURE FOR EACH INDIVIDUAL
// ----------------------------------------------------------------------------------

void Output::entropy(Param &p, LCAParam &lp, Data &d)
{

  if (lp.nClass >1)
    {
      cout << "\nX:"<<p.tag<<": Assignment entropy";
      cout << "\nX:"<<p.tag<<":\tID\tE";
      
      for (int i=1; i <= d.nInd; i++)
	{
	  cout << "\nE:"<<p.tag<<":\t" << d.id(i);
	  
	  CColVector E(lp.nClass), E2(lp.nClass);
	  
	  for (int j=1; j <= lp.nClass; j++)
	    {
	      E[j] = h[j][d.ic(i)];
	      if (j==1) E2[j] = 1; else E2[j] = 0;
	    }
	  
	  double entropy = 0 ; 
	  for (int j=1; j <= lp.nClass; j++)
	    if (h[j][d.ic(i)] > 0)
	      entropy -= h[j][d.ic(i)] 
		* ( log(h[j][d.ic(i)])/log((double)lp.nClass));
	  if (entropy <= 0) entropy = 0 ;
	  
	  cout << "\t" 
	       << entropy;
//  	       << mymath::variance(E) / mymath::variance(E2) 
//  	       << "\t"

	}
    }
}

void Output::locinfo(Param &p, LCAParam &lp, Data &d)
{
  // Calculate average genetic distance for 1 locus 
  // across all pairs of different classes

  if (lp.nClass > 1)
    {
      cout << "\n\nX:"<<p.tag<<": Inter-class locus variation";
      for ( int l = 1 ; l <= lp.nLoci ; l++ )
	{
	  cout << "\nL:"<<p.tag<<":\t"<<l<<"\t";
	  
	  double dis = 0 ;
	  
	  for (int i = 1 ; i <= lp.nClass; i++)
	    {
	      for (int j = 1 ; j <= i ; j++)
		{
		  double dis2 = 1;
		  for (int a = 3 ; a <= d.nallele(l); a++)
		    dis2 -= sqrt(r[l][i][a] * r[l][j][a]);
		  
		  if (dis2>0) dis += sqrt(dis2);
		}
	    }
	  dis /= (float)((lp.nClass * (lp.nClass-1))/2);
	  cout << dis;
	}
    }
}



// ----------------------------------------------------------------------------------
// CALCULATE AND OUTPUT INTER-CLASS GENETIC DISTANCE MEASURE 
// ----------------------------------------------------------------------------------


void Output::cslocinfo(Param &p, LCAParam &lp, Data &d)
{
  // Calculate average genetic distance for 1 locus 
  // across one class versus all others

  if (lp.nClass > 1)
    {
      // Header
      cout << "\n\nS:"<<p.tag<<": Class-specific locus variation"
	   << "\nS:"<<p.tag<<":\t";
      for (int i = 1 ; i <= lp.nClass; i++)
	cout <<"\t"<<i;

      // For eacj locus l
      for ( int l = 1 ; l <= lp.nLoci ; l++ )
	{
	  cout << "\nS:"<<p.tag<<":\t"<<l;
	  
	  // For each class '1'	  
	  for (int j = 1 ; j <= lp.nClass; j++)
	    {

	      double dis = 0 ;

	      // For each 'other' class
	      for (int j2 = 1 ; j2 <= lp.nClass; j2++)
		{
		  if (j!=j2)
		    {
		      double dis2 = 1;
		      for (int a = 3 ; a <= d.nallele(l); a++)
			dis2 -= sqrt(r[l][j][a] * r[l][j2][a]);
		      if (dis2>0) dis += sqrt(dis2);
		    }
		}
	      dis /= (float)(lp.nClass-1);
	      cout << "\t" << dis;
	    }
	}
    }
}



// ----------------------------------------------------------------------------------
// CALCULATE AND OUTPUT INTER-CLASS GENETIC DISTANCE MEASURE 
// ----------------------------------------------------------------------------------

void Output::distance(Param &p, LCAParam &lp, Data &d)
{

  if (lp.nClass > 1)
    {
      cout << "\n\nD:"<<p.tag<<": Inter-class Nei genetic distance matrix"
	   << "\nD:"<<p.tag<<":\t";
      
      for (int i = 1 ; i <= lp.nClass; i++)
	cout <<"\t"<<i;
      
      for (int i = 1 ; i <= lp.nClass; i++)
	{
	  cout << "\nD:"<<p.tag<<":\t"<<i;
	  for (int j = 1 ; j <= lp.nClass ; j++)
	    {

	      // Original version
	      double dis = 0 ;
	      for ( int l = 1 ; l <= lp.nLoci ; l++ )
		{
		  double dis2 = 1;
		  for (int a = 3 ; a <= d.nallele(l); a++)
		    dis2 -= sqrt(r[l][i][a] * r[l][j][a]);
		  
		  if (dis2>0) dis += sqrt(dis2);
		}		
	      // cout << "\t"<< dis / (float)lp.nLoci ; 

	      // Second implementation
	      dis = 0;
	      // Sum of locus, allele PaPb
	      double numerator = 0;
	      for ( int l = 1 ; l <= lp.nLoci ; l++ )
		for ( int a = 3 ; a <= d.nallele(l); a++)
		  numerator += r[l][i][a] * r[l][j][a];
	      numerator /= lp.nLoci;
	      
	      double denominator_1 = 0;
	      double denominator_2 = 0; 
	      for ( int l = 1 ; l <= lp.nLoci ; l++ )
		for ( int a = 3 ; a <= d.nallele(l); a++)
		  {
		    denominator_1 += r[l][i][a] * r[l][i][a];
		    denominator_2 += r[l][j][a] * r[l][j][a];
		  }
	      denominator_1 /= lp.nLoci;
	      denominator_2 /= lp.nLoci;
	      double denominator = sqrt( denominator_1 * denominator_2);
	      double nei = -log(numerator/denominator);
	      if (nei <= 0) nei = 0; 
	      cout << "\t" << nei;
	    }
	}
      cout << "\n";
    }
}
