
#ifndef __PARAM_H__
#define __PARAM_H__

#include <string>
#include "shaunlib.h"

class Data;
class LCAParam;

// ----------------------------------------------------------------------------------
// DEFINE AND DEFAULT VALUES FOR GENERAL PARARMETERS
// ----------------------------------------------------------------------------------

class Param 
{
 public:
  Param();

  void get_param(LCAParam &lp, Data &d);
  void get_loci_ind(LCAParam &lp, Data &d);
  void get_class(LCAParam &lp);
  void get_storage(LCAParam &lp, Data &d);

  string            title;
  string            tag;
  string            datafile;         // default data file
  bool              header_row;

  bool              haploid;          // All haploid
  bool              structure_format; // F = own, T = Structure
  bool              fix_to_class;     // Fix to class column ?
  bool              fix_n_loci; 
  bool              group_geno;       // analyse individual-by-individual ?
  bool              entropy;          // output this measure for inds
  bool              locinfo;          // locus information
  bool              cslocinfo;        // class-specifc locus information
  bool              verbose1; 
  bool              verbose2;
  bool              verbose3;
  bool              min; 
  bool              debug;

  bool              distance;
  bool              xdata;            // any X chr data?
  bool              hwd;              // allow HWD by coding genotypes
                                      // as alleles, then analysing 
                                      // as haploid

  int               nInd;             // # individuals, calculated from file
  int               nLoci;            // # loci, calculated from file
  int               nPheno;           // default no phenotypes
  int               nAllele;          // defined in parameter set, default to 5
  int               valMissing;       // missing allele code
  int               valX;             // X chr missing code 
  int               admix_res;        // Admixture resolution
  long              random_seed;      // random number seed
  int               nItRepeat;        // number of repeat iterations
  double            tol;              // E-M convergence tolerance

};


// ----------------------------------------------------------------------------------
// LATENT CLASS ANALYSIS VARIABLES
// ----------------------------------------------------------------------------------

class LCAParam
{
 public:

  LCAParam();
  ~LCAParam();

  void make_admix();
  void free_storage();

  int               nClass;             // Number of Classses
  int               nPureClass;         // NUmber of pure ancestral classes
  int               nAdmixClass;        // Number of ancestral classes to be admixed
  float             nAdmixClass2;       // Derived number of admixed classes
  int               nAncClass;          // = nPureClass + nAdmixClass

  int               admix_res;          // Admixture resolution

  bool              setPure;            // True when CLASS has been used
  bool              setAuto;
  bool              setA1;

  int               nAdmix1Class;       // Number of specified admix1 classes
  int               Admix1Class1[19];   // Hard coded for up to 20 types of 
  int               Admix1Class2[19];   // this new class

  int               nAdmix2Class;       // Number of new admixed classes
  int               Admix2Class1[19];   // Hard coded for up to 20 types of 
  int               Admix2Class2[19];   // this new class

  CMatrix         h;                  // posterior class probabilities ( [class][config] )
  IntArray        indClass;           // 'best-fit' class

  CVector         n;                  // class-specific individual frequencies
  CVector         n_;                 // class-specific individual counts

  CMatrix         anc_n_;             // ancestral class-specific individual counts

  CTensor         r_;                 // [ind][loc][allele]
  CTensor         r;                  // class-specific allele frequencies 

  CTensor         anc_r_;             // ancestral class allele counts
  CTensor         anc_r;              // ancestral class allele frequencies

  CVector         contrib;            // 
  CVector         f;                  // likelihood of each invidual 
  CMatrix         cls;                // weighting definition matrix 
                                        // for ancestral to class mapping

  int               nAllele; 
  int               nLoci;

  int               npar;
 
};


#endif
